@@ -23,6 +23,35 @@ from docker.types import LogConfig
 
 
 class AgentLifeCycle:
+    """
+    A class to manage agent and Cognicept's docker container lifecycle
+    ...
+
+    Parameters
+    ----------
+    None
+
+    Methods
+    -------
+    configure_containers(cfg):
+        Loads agent configuration from `COG_AGENT_CONTAINERS` and `COG_AGENT_IMAGES`
+    get_status(args):
+        Prints status of docker containers listed in `COG_AGENT_CONTAINERS`.
+    get_last_event(args):
+        Prints last log in `~/.cognicept/agent/logs/`.
+    restart(args):
+        Stops and starts the containers listed in `COG_AGENT_CONTAINERS`.
+    start(args):
+        Starts the containers listed in `COG_AGENT_CONTAINERS` and datadog. If `args` has parameter `list`, starts only containers in the list.
+        If `args` has parameter `datadog`/`agents` it will start only datadog/containers.
+    stop(args):
+        Stops the containers listed in `COG_AGENT_CONTAINERS` and datadog. If `args` has parameter `list`, stops only containers in the list.
+        If `args` has parameter `datadog`/`agents` it will stop only datadog/containers.
+    update(args):
+        Pulls docker images listed in `COG_AGENT_IMAGES`.
+    run_orbitty(args):
+        Starts Orbitty.
+    """
 
     # default configuration of containers and images
     _docker_container_names = ["cgs_diagnostics_agent", "remote_intervention_agent",
@@ -36,6 +65,16 @@ class AgentLifeCycle:
     _docker_images["cgs_bagger_server"] = "412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/cognicept_rosbagger:latest"
 
     def configure_containers(object, cfg):
+        """
+        Loads agent configuration from `COG_AGENT_CONTAINERS` and `COG_AGENT_IMAGES`
+
+                Parameters:
+                        cfg (Configuration): Cognicept configuration
+                Returns:
+                        None
+        """
+
+
         if("COG_AGENT_CONTAINERS" in cfg.config and "COG_AGENT_IMAGES" in cfg.config):
             container_names = cfg.config["COG_AGENT_CONTAINERS"].split(";")
             image_names = cfg.config["COG_AGENT_IMAGES"].split(";")
@@ -57,7 +96,15 @@ class AgentLifeCycle:
             print(
                 "Undefined `COG_AGENT_CONTAINERS` or `COG_AGENT_IMAGES`. Using default.")
 
-    def get_latest_log_loc(object, args):
+    def _get_latest_log_loc(object, args):
+        """
+        Retrieve path to the last log in `~/.cognicept/agent/logs/` relative to `~/.cognicept/` or `path` specified by args.
+
+                Parameters:
+                        args: populated argument namespace returned by `argparse.parse_args()`
+                Returns:
+                        latest_log_loc (str): path to latest log relative to `~/.cognicept/`
+        """
         # get latest log location
         latest_log_loc_file_path = os.path.expanduser(
             args.path+"agent/logs/latest_log_loc.txt")
@@ -75,10 +122,17 @@ class AgentLifeCycle:
         return latest_log_loc
 
     def get_status(object, args):
+        """
+        Prints status of docker containers listed in `COG_AGENT_CONTAINERS`.
+
+                Parameters:
+                        args: populated argument namespace returned by `argparse.parse_args()`
+                
+        """
         client = docker.from_env()
         # check status of cgs_agent
         # get latest log location
-        latest_log_loc = object.get_latest_log_loc(args)
+        latest_log_loc = object._get_latest_log_loc(args)
         # read latest status and set for display
         file_path = os.path.expanduser(
             args.path+latest_log_loc+"/logDataStatus.json")
@@ -106,7 +160,7 @@ class AgentLifeCycle:
                     if(container_name == "cgs_agent"):
                         print(cgs_agent_status)
                     elif(container_name == "remote_intervention_agent"):
-                        object.parseRemoteInterventionAgentLogs(
+                        object._parse_remote_intervention_agent_logs(
                             container.logs(tail=50))
                     elif(container_name == "cgs_diagnostics_agent"):
                         print(cgs_agent_status)
@@ -116,9 +170,15 @@ class AgentLifeCycle:
                 print(bcolors.FAIL + "CONTAINER NOT FOUND" + bcolors.ENDC)
 
     def get_last_event(object, args):
+        """
+        Prints last log in `~/.cognicept/agent/logs/`.
 
+                Parameters:
+                        args: populated argument namespace returned by `argparse.parse_args()`
+                
+        """
         # get latest log location
-        latest_log_loc = object.get_latest_log_loc(args)
+        latest_log_loc = object._get_latest_log_loc(args)
 
         # read and display latest log if any
         file_path = os.path.expanduser(args.path+latest_log_loc)
@@ -137,8 +197,14 @@ class AgentLifeCycle:
         except:
             print(bcolors.WARNING + "No event logs present." + bcolors.ENDC)
 
-    def parseRemoteInterventionAgentLogs(object, logs):
-        # Parses logs to find status of remote intervention agent
+    def _parse_remote_intervention_agent_logs(object, logs):
+        """
+        Parses logs to find status of remote intervention agent. Prints status.
+
+                Parameters:
+                        logs: container logs
+                
+        """
         logs_lines = logs.splitlines()
         # parse logs to get current status
         ri_agent_status = {}
@@ -171,11 +237,28 @@ class AgentLifeCycle:
         print(output_text)
 
     def restart(object, args):
+        """
+        Stops and starts the containers listed in `COG_AGENT_CONTAINERS`.
+
+                Parameters:
+                        args: populated argument namespace returned by `argparse.parse_args()`
+                Returns:
+                        result (bool): True if succeeded
+        """
         object.stop(args)
         result = object.start(args)
         return result
 
     def start(object, args):
+        """
+        Starts the containers listed in `COG_AGENT_CONTAINERS` and datadog. If `args` has parameter `list`, starts only containers in the list.
+        If `args` has parameter `datadog`/`agents` it will start only datadog/containers.
+
+                Parameters:
+                        args: populated argument namespace returned by `argparse.parse_args()`
+                Returns:
+                        result (bool): True if succeeded
+        """
         if(args.agents or (hasattr(args, 'list') and len(args.list) > 0)):
             print("Starting agents")
             result = object.run_agents(args)
@@ -192,7 +275,15 @@ class AgentLifeCycle:
         return result
 
     def stop(object, args):
+        """
+        Stops the containers listed in `COG_AGENT_CONTAINERS` and datadog. If `args` has parameter `list`, stops only containers in the list.
+        If `args` has parameter `datadog`/`agents` it will stop only datadog/containers.
 
+                Parameters:
+                        args: populated argument namespace returned by `argparse.parse_args()`    
+                Returns:
+                        result (bool): True if succeeded            
+        """
         if(args.agents or (hasattr(args, 'list') and len(args.list) > 0)):
             print("Stopping agents")
             result = object.remove_agents(args)
@@ -213,6 +304,14 @@ class AgentLifeCycle:
         return True
 
     def remove_agents(object, args):
+        """
+        Stops the containers listed in `args.list`.
+
+                Parameters:
+                        args: populated argument namespace returned by `argparse.parse_args()`    
+                Returns:
+                        result (bool): True if succeeded            
+        """
 
         # if list of agents is not specified, restart all
         if(not hasattr(args, 'list') or len(args.list) == 0):
@@ -237,7 +336,14 @@ class AgentLifeCycle:
         return flag_success
 
     def run_agents(object, args):
+        """
+        Starts the containers listed in `args.list`.
 
+                Parameters:
+                        args: populated argument namespace returned by `argparse.parse_args()`    
+                Returns:
+                        result (bool): True if succeeded            
+        """
         # if list of agents is not specified, restart all
         if(not hasattr(args, 'list') or len(args.list) == 0):
             args.list = object._docker_container_names
@@ -311,18 +417,42 @@ class AgentLifeCycle:
         return success_flag
 
     def stop_datadog(object, args):
+        """
+        Stops datadog.
+
+                Parameters:
+                        args: populated argument namespace returned by `argparse.parse_args()`    
+                Returns:
+                        result (bool): True if succeeded
+        """
         sts = subprocess.Popen(
             "systemctl stop datadog-agent", shell=True).wait()
         print("Health monitor agent stopped")
         return True
 
     def run_datadog(object, args):
+        """
+        Starts datadog.
+
+                Parameters:
+                        args: populated argument namespace returned by `argparse.parse_args()`    
+                Returns:
+                        result (bool): True if succeeded
+        """
         sts = subprocess.Popen(
             "systemctl start datadog-agent", shell=True).wait()
         print("Health monitor agent started")
         return True
 
     def status_datadog(object, args):
+        """
+        Prints datadog status.
+
+                Parameters:
+                        args: populated argument namespace returned by `argparse.parse_args()`    
+                Returns:
+                        result (bool): True if succeeded
+        """
         sts = subprocess.Popen(
             ["systemctl status datadog-agent", "-p"], shell=True, stdout=DEVNULL, stderr=DEVNULL).wait()
         if sts == 0:
@@ -336,6 +466,14 @@ class AgentLifeCycle:
                   bcolors.WARNING + " INACTIVE" + bcolors.ENDC)
 
     def run_orbitty(object, args):
+        """
+        Starts Orbitty.
+
+                Parameters:
+                        args: populated argument namespace returned by `argparse.parse_args()`    
+                Returns:
+                        None           
+        """
         os.system("xhost +local:root")
         client = docker.from_env()
         try:
@@ -367,17 +505,31 @@ class AgentLifeCycle:
         os.system("xhost -local:root")
 
     def update(object, args):
+        """
+        Pulls docker images listed in `COG_AGENT_IMAGES`.
 
+                Parameters:
+                        args: populated argument namespace returned by `argparse.parse_args()`    
+                Returns:
+                        result (bool): True if succeeded           
+        """
         # Attempt to login to ECR client
-        login_success = object.ecr_login(args)
+        login_success = object._ecr_login(args)
 
         # If login failed, return false
         if login_success is False:
-            print("You don't seem to have ECR repository permissions. Check your AWS credentials with the Cognicept team.")
+            print("Your update credentials may have expired. Run `cognicept keyrotate` to refresh credentials and try `cognicept update` again.")
             return False
 
         print("Info: This may take a while depending on your connection.")
         images = set(object._docker_images.values())
+
+        # load extra images to update
+        if("COG_EXTRA_IMAGES" in args.config.config):
+            image_names = args.config.config["COG_EXTRA_IMAGES"].split(";")
+            if(len(image_names) > 0):
+                images = images.union(set(image_names))
+
         N = len(images)
         i = 0
         success_flag = True
@@ -415,33 +567,46 @@ class AgentLifeCycle:
         print("Info: Run `cognicept restart` to redeploy updated agents.")
         return success_flag
 
-    def construct_ecr_client(object, config_obj):
+    def _construct_ecr_client(object, config_obj):
         """
-        Utility function to create ECR client based on runtime.env credentials
+        Member utility function to create ECR client `object.ecr_client` based on runtime.env credentials
+
+                Parameters:
+                        config_obj (Configuration): object holding Cognicept configuration
+                Returns:
+                        None      
         """
         # Get config
-        local_cfg = config_obj.config
-
-        if 'AWS_SESSION_TOKEN' in local_cfg:
+        local_cfg = config_obj.fetch_aws_keys()
+        if local_cfg==False:
+            return False
+        if 'SessionToken' in local_cfg:
             object.ecr_client = boto3.client(
                 'ecr', region_name='ap-southeast-1',
-                aws_access_key_id=local_cfg['AWS_ACCESS_KEY_ID'],
-                aws_secret_access_key=local_cfg['AWS_SECRET_ACCESS_KEY'],
-                aws_session_token=local_cfg['AWS_SESSION_TOKEN'])
+                aws_access_key_id=local_cfg['AccessKeyId'],
+                aws_secret_access_key=local_cfg['SecretAccessKey'],
+                aws_session_token=local_cfg['SessionToken'])
         else:
             object.ecr_client = boto3.client(
                 'ecr', region_name='ap-southeast-1',
-                aws_access_key_id=local_cfg['AWS_ACCESS_KEY_ID'],
-                aws_secret_access_key=local_cfg['AWS_SECRET_ACCESS_KEY'])
+                aws_access_key_id=local_cfg['AccessKeyId'],
+                aws_secret_access_key=local_cfg['SecretAccessKey'])
+        return True
 
-    def ecr_login(object, args):
+    def _ecr_login(object, args):
         """
-        Utility function that is called to login to ECR
+        Member utility function that is called to login to ECR
+
+                Parameters:
+                        args: populated argument namespace returned by `argparse.parse_args()`
+                Returns:
+                        result (bool): True if succeeded
         """
         # Construct client
-        object.construct_ecr_client(args.config)
+        result = object._construct_ecr_client(args.config)
+        if result==False:
+            return False
         num_retries = 3
-
         for trial in range(num_retries):
             try:
                 # Get token
